/*****************************************************************************
*
*      exser.h		XR16L788/784 Family Multiport Serial Driver.
*
*      Copyright (C) 2001  Exar Corp.	(www.exar.com)
*	   Copyright (C) 1991, 1992, Linus Torvalds
*      Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997,
*					 1998, 1999, Theodore Ts'o
*
*      This code is loosely based on the Linux serial driver, written by
*      Linus Torvalds, Theodore T'so and others.
*
*      This program is free software; you can redistribute it and/or modify
*      it under the terms of the GNU General Public License as published by
*      the Free Software Foundation; either version 2 of the License, or
*      (at your option) any later version.
*
*      This program is distributed in the hope that it will be useful,
*      but WITHOUT ANY WARRANTY; without even the implied warranty of
*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*      GNU General Public License for more details.
*
*      You should have received a copy of the GNU General Public License
*      along with this program; if not, write to the Free Software
*      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
*	   XR16L788/784 Family Multiport Serial Driver
*	   for			: LINUX		(Tested with RedHat 7.0 and 7.1)
*	   date			: August, 2001
*	   version		: 1.1
********************************************************************************/

#define	VERSION_CODE(ver, rel, seq)	((ver << 16) | (rel << 8) | seq)
#ifdef CONFIG_PCI
#include <linux/pci.h>
#endif

//if PCI device is used, turn it on.
//#define XR154_PCI
//#define XR158_PCI
//command line installation (insmod name.o IRQ=8 IO=0x600)

static unsigned short IRQ = 5;
#if !defined(XR158_PCI) && !defined(XR154_PCI)
static int IO = 0x500;
#endif
MODULE_PARM(IRQ, "i");
#if !defined(XR158_PCI) && !defined(XR154_PCI)
MODULE_PARM(IO, "i");
#endif

//queue for BH irq processing
DECLARE_TASK_QUEUE(tq_serial);

#define         EXSER_VERSION           "1.2"
#define         EXSERMAJOR              30
#define         EXSERCUMAJOR            35

#ifdef CONFIG_PCI
#if (LINUX_VERSION_CODE < VERSION_CODE(2, 1, 0))
#include <linux/bios32.h>
#endif
#include <linux/pci.h>
#endif 

#if (LINUX_VERSION_CODE < VERSION_CODE(2, 1, 0))
#define copy_from_user	memcpy_fromfs
#define copy_to_user	memcpy_tofs
#define put_to_user(arg1, arg2)		put_fs_long(arg1, (unsigned long *) arg2)
#define get_from_user(arg1, arg2)	arg1 = get_fs_long((unsigned long *) arg2)
#define schedule_timeout(x) {current->timeout = jiffies + (x); schedule();}
#define signal_pending(x) ((x)->signal & ~(x)->blocked)
#else
#include <asm/uaccess.h>
#define put_to_user(arg1, arg2)		put_user(arg1, (unsigned long *) arg2)
#define get_from_user(arg1, arg2)	get_user(arg1, (unsigned int *) arg2)
#endif 

#define EXSER_EVENT_TXLOW       1
#define EXSER_EVENT_HANGUP      2
#define SERIAL_DO_RESTART
#define EXSER_BOARDS            4       //Max. boards
#define EXSER_PORTS             32      //Max. ports
#define EXSER_PORTS_PER_BOARD   8       //Max. ports per board
#define EXSER_ISR_PASS_LIMIT    256
#define EXSER_ERR_IOADDR        -1
#define EXSER_ERR_IRQ           -2
#define EXSER_ERR_IRQ_CONFLIT   -3
#define EXSER_ERR_VECTOR        -4
#define SERIAL_TYPE_NORMAL		1
#define SERIAL_TYPE_CALLOUT		2
#define WAKEUP_CHARS			256
#define UART_MCR_AFE			0x20
#define UART_LSR_SPECIAL		0x1E
#define C158_ASIC_ID			1
#define C104_ASIC_ID			2
#define CI104J_ASIC_ID			5
#define ISA_XR_DVID				0x8D
#define ISA_XR788				0x28
#define ISA_XR784				0x24
#define LED_ADDR				0x93
#define PORTNO(x)		(MINOR((x)->device) - (x)->driver.minor_start)
#define RELEVANT_IFLAG(iflag)	(iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))
#define IRQ_T(info) ((info->flags & ASYNC_SHARE_IRQ) ? SA_SHIRQ : SA_INTERRUPT)
#define _INLINE_	inline

#ifndef MIN
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#endif
#ifndef PCI_VENDOR_ID_XR
#define PCI_VENDOR_ID_XR		0x13a8
#endif
#ifndef PCI_DEVICE_ID_XR158
#define PCI_DEVICE_ID_XR158      0x158
#endif
#ifndef PCI_DEVICE_ID_XR154
#define PCI_DEVICE_ID_XR154      0x154	
#endif
#ifndef PCI_DEVICE_ID_C152
#define PCI_DEVICE_ID_C152      0x152
#endif
#ifndef PCI_DEVICE_ID_C104
#define PCI_DEVICE_ID_C104		0x1040
#endif

enum	{
		ISA_BOARD_XR788   = 1,
        ISA_BOARD_XR784,
        EXSER_BOARD_CI104J,
        PCI_BOARD_XR158,
        PCI_BOARD_XR154,	
        EXSER_BOARD_C152_PCI,	
        EXSER_BOARD_C104_PCI,
};

static char *exser_brdname[] = {
		"XR158/154 series",
		"XR788/784 series",
        "CI-104J series",
        "C158/PCI series",
        "C154/PCI series",		
        "C152/PCI series",		
        "C104/PCI series",
};

static int exser_numports[] = {
		8,	
		4,
        4,
		8,
		4,	
		2,	
		4,
};
	
//EXAR ioctls
#define		EXAR				  0x400
#define     EXAR_GETDATACOUNT     (EXAR + 23)
#define     EXAR_GET_CONF         (EXAR + 35)
#define     EXAR_DIAGNOSE         (EXAR + 50)
#define     EXAR_CHKPORTENABLE    (EXAR + 60)
#define     EXAR_HighSpeedOn      (EXAR + 61)
#define     EXAR_GET_MAJOR        (EXAR + 63)
#define     EXAR_GET_CUMAJOR      (EXAR + 64)
#define     EXAR_GETMSTATUS       (EXAR + 65)
#define		EXAR_READ_REG         (EXAR + 66)
#define 	EXAR_WRITE_REG        (EXAR + 67)
#define 	EXAR_SET_RS485        (EXAR + 68)

#ifdef CONFIG_PCI
typedef struct {
	unsigned short	vendor_id;
	unsigned short	device_id;
	unsigned short	board_type;
} exser_pciinfo;
static exser_pciinfo    exser_pcibrds[] = {
        {PCI_VENDOR_ID_XR, PCI_DEVICE_ID_XR158, PCI_BOARD_XR158},
        {PCI_VENDOR_ID_XR, PCI_DEVICE_ID_XR154, PCI_BOARD_XR154},
};
typedef struct _exar_pci_info {
	unsigned short busNum;
	unsigned short devNum;
} exar_pci_info;
#endif //CONFIG_PCI

static int ioaddr[EXSER_BOARDS] = {0, 0, 0, 0};
static int ttymajor = EXSERMAJOR;
static int calloutmajor = EXSERCUMAJOR;
static int verbose = 1;

//Variables for insmod
MODULE_AUTHOR("EXAR");
MODULE_DESCRIPTION("EXAR Multiport Board Device Driver");
MODULE_PARM(ioaddr,			"1-4i");
MODULE_PARM(ttymajor,       "i");
MODULE_PARM(calloutmajor,   "i");
MODULE_PARM(verbose,        "i");

struct exser_hwconf {
	int			board_type;
	int			ports;
	int			irq;
	int			vector;
	int			vector_mask;
	int			uart_type;
    int         ioaddr[EXSER_PORTS_PER_BOARD];
    int         baud_base[EXSER_PORTS_PER_BOARD];
	int			brdid;
    exar_pci_info    pciInfo;
};

struct exser_struct {
	int			port;
	int			base;			//port base address
	int			irq;			//port using irq no.
	int			vector; 		//port irq vector
	int			vectormask;		//port vector mask
	int			rx_trigger;		//Rx fifo trigger level
	int			baud_base;		//max. speed
	int			flags;			//defined in tty.h
	int			type;			//UART type
	struct tty_struct *	tty;
	int			read_status_mask;
	int			ignore_status_mask;
	int			xmit_fifo_size;
	int			custom_divisor;
	int			x_char; 		//xon/xoff character
	int			close_delay;
	unsigned short		closing_wait;
	int			IER;			//Interrupt Enable Register
	int			MCR;			//Modem control register
	unsigned long		event;
	int			count;			//# of fd on device
	int			blocked_open;	//# of blocked opens
	long			session;	//Session of opening process
	long			pgrp;		//pgrp of opening process
	unsigned char		*xmit_buf;
	int			xmit_head;
	int			xmit_tail;
	int			xmit_cnt;
	struct tq_struct	tqueue;
	struct termios		normal_termios;
	struct termios		callout_termios;
	wait_queue_head_t	open_wait;
	wait_queue_head_t	close_wait;
	wait_queue_head_t	delta_msr_wait;
	struct async_icount	icount; 	//kernel counters for the 4 input interrupts
};

struct exser_log {
	int	    tick;
    int     rxcnt[EXSER_PORTS];
    int     txcnt[EXSER_PORTS];
};

struct exser_mstatus{
    tcflag_t  cflag;
    int  cts;
    int  dsr;
    int  ri;
    int  dcd;
};

static  struct exser_mstatus GMStatus[EXSER_PORTS];

struct exser_read_reg {
	int reg;
	int regvalue;
};

static struct tty_driver        exvar_sdriver, exvar_cdriver;
static int                      exvar_refcount;
static struct exser_struct      exvar_table[EXSER_PORTS];
static struct tty_struct *      exvar_tty[EXSER_PORTS+1];
static struct termios *         exvar_termios[EXSER_PORTS+1];
static struct termios *         exvar_termios_locked[EXSER_PORTS+1];
static struct exser_log         exvar_log;
static int                      exvar_diagflag;
/*
 * exvar_tmp_buf is used as a temporary buffer by serial_write. We need
 * to lock it in case the memcpy_fromfs blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */

//Changed to support the 2.4 kernel
static unsigned char *	exvar_tmp_buf;
static DECLARE_MUTEX(exvar_tmp_buf_sem);

//This is used to figure out the divisor speeds and the timeouts
static int exvar_baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600, 0 };
	
struct exser_hwconf exsercfg[EXSER_BOARDS];

//static functions:
#if( LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0) )
static int	exar_init_module(void);
static void	exar_cleanup_module(void);
module_init( exar_init_module );
module_exit( exar_cleanup_module );
#endif

static void     exser_getcfg(int, struct exser_hwconf *);
static int		exser_initbrd(int, struct exser_hwconf *);
static int      exser_init(void);
static int      exser_get_PCI_conf(int ,int ,int ,struct exser_hwconf *, struct pci_dev *);
static void     exser_do_softint(void *);
static int      exser_open(struct tty_struct *, struct file *);
static void     exser_close(struct tty_struct *, struct file *);
static int      exser_write(struct tty_struct *, int, const unsigned char *, int);
static int      exser_write_room(struct tty_struct *);
static void     exser_flush_buffer(struct tty_struct *);
static int      exser_chars_in_buffer(struct tty_struct *);
static void     exser_flush_chars(struct tty_struct *);
static void     exser_put_char(struct tty_struct *, unsigned char);
static int      exser_ioctl(struct tty_struct *, struct file *, uint, ulong);
static int      exser_ioctl_special(unsigned int, unsigned long);
static void     exser_throttle(struct tty_struct *);
static void     exser_unthrottle(struct tty_struct *);
static void     exser_set_termios(struct tty_struct *, struct termios *);
static void     exser_stop(struct tty_struct *);
static void     exser_start(struct tty_struct *);
static void     exser_hangup(struct tty_struct *);
static void     exser_interrupt(int, void *, struct pt_regs *);
static _INLINE_ void exser_receive_chars(struct exser_struct *, int *);
static _INLINE_ void exser_transmit_chars(struct exser_struct *);
static _INLINE_ void exser_check_modem_status(struct exser_struct *, int);
static int      exser_block_til_ready(struct tty_struct *, struct file *, struct exser_struct *);
static int      exser_startup(struct exser_struct *);
static void     exser_shutdown(struct exser_struct *);
static int      exser_change_speed(struct exser_struct *, struct termios *);
static int      exser_get_serial_info(struct exser_struct *, struct serial_struct *);
static int      exser_set_serial_info(struct exser_struct *, struct serial_struct *);
static int      exser_get_lsr_info(struct exser_struct *, unsigned int *);
static void     exser_send_break(struct exser_struct *, int);
static int      exser_get_modem_info(struct exser_struct *, unsigned int *);
static int      exser_set_modem_info(struct exser_struct *, unsigned int, unsigned int *);
static void		exser_setled(struct exser_hwconf *);
static void		exser_clearled(struct exser_hwconf *);
static int		exser_read_reg(struct exser_struct *, struct exser_read_reg *);
static int		exser_write_reg(struct exser_struct *, struct exser_read_reg *);
static int		exser_set_autoRS485(struct exser_struct *, unsigned int);
static unsigned char	byte_read( unsigned long);
static void		byte_write( unsigned char, unsigned long);
static int		get_isr_conf(int, struct exser_hwconf *);




















